import inspect
from typing import (
    Any,
    Callable,
    Generic,
    List,
    Optional,
    Protocol,
    TypeVar,
    Union,
    runtime_checkable,
)

from fastapi import APIRouter, HTTPException, Request, Response
from fastapi.routing import APIRoute
from pydantic import BaseModel

from fastapi_resources.resources.sqlmodel.exceptions import NotFound
from fastapi_resources.resources.types import ResourceProtocol
from fastapi_resources.routers import decorators


class TCreatePayload(BaseModel):
    pass


class TUpdatePayload(BaseModel):
    pass


TResource = TypeVar("TResource", bound=ResourceProtocol)


@runtime_checkable
class Action(Protocol):
    detail: bool
    methods: decorators.methods
    kwargs: dict[str, Any]

    def __call__(self):
        ...


class ResourceRoute(APIRoute):
    def get_route_handler(self) -> Callable:
        original_route_handler = super().get_route_handler()

        async def custom_route_handler(request: Request) -> Response:
            response: Response = await original_route_handler(request)

            if resource := getattr(request, "resource", None):
                resource.close()

            return response

        return custom_route_handler


class ResourceRouter(APIRouter, Generic[TResource]):
    resource_class: type[TResource]

    def __init__(
        self,
        *,
        resource_class: Optional[type[TResource]] = None,
        **kwargs,
    ) -> None:
        super().__init__(route_class=ResourceRoute, **kwargs)

        if resource_class:
            self.resource_class = resource_class

        self.method_replacements = self.get_method_replacements()
        self.ReadResponseModel = self.get_read_response_model()
        self.ListResponseModel = self.get_list_response_model()

        self._patch_route_types()
        self._link_routes()

    def get_read_response_model(self):
        return self.resource_class.Read

    def get_list_response_model(self):
        Read = self.resource_class.Read
        return List[Read]

    def get_update_model(self):
        return self.resource_class.Update

    def get_create_model(self):
        return self.resource_class.Create

    def get_method_replacements(self):
        return {
            "_create": {TCreatePayload: self.get_create_model()},
            "_update": {TUpdatePayload: self.get_update_model()},
        }

    def api_route(self, *args, **kwargs):
        try:
            return super().api_route(*args, **kwargs)
        except NotFound as e:
            raise HTTPException(status_code=404, detail=str(e))

    def _patch_route_types(self):
        for method_name, replacements in self.method_replacements.items():
            method_instance = getattr(self, method_name, None)

            if not method_instance:
                continue

            original_signature = inspect.signature(method_instance)
            updated_params = [
                inspect.Parameter(
                    name=param.name,
                    kind=param.kind,
                    default=param.default,
                    annotation=replacements.get(param.annotation, param.annotation),
                )
                for param in original_signature.parameters.values()
            ]
            new_params = [
                inspect.Parameter(
                    name=key.replace("*", ""),
                    kind=inspect._ParameterKind.KEYWORD_ONLY,
                    default=None,
                    annotation=annotation,
                )
                for key, annotation in replacements.items()
                if isinstance(key, str) and key.startswith("*")
            ]
            updated_signature = inspect.Signature(updated_params + new_params)

            # Required to avoid closing over method_name
            def factory(_method_name):
                def wrapper(*args, **kwargs):
                    class_method = getattr(self.__class__, _method_name)
                    return class_method(self, *args, **kwargs)

                return wrapper

            setattr(self, method_name, factory(method_name))
            getattr(self, method_name).__signature__ = updated_signature

    def _link_routes(self):
        resource_class = self.resource_class

        self._link_actions()

        if resource_class.retrieve:
            self.get(
                f"/{{id}}",
                response_model=self.ReadResponseModel,
                response_model_exclude_unset=True,
                summary=f"Get {resource_class.name}",
            )(self._retrieve)

        if resource_class.list:
            self.get(
                f"",
                response_model=self.ListResponseModel,
                response_model_exclude_unset=True,
                summary=f"Get {resource_class.name} list",
            )(self._list)

        if resource_class.create:
            self.post(
                f"",
                response_model=self.ReadResponseModel,
                response_model_exclude_unset=True,
                summary=f"Create {resource_class.name}",
                status_code=201,
            )(self._create)

        if resource_class.update:
            self.patch(
                f"/{{id}}",
                response_model=self.ReadResponseModel,
                response_model_exclude_unset=True,
                summary=f"Update {resource_class.name}",
            )(self._update)

        if resource_class.delete:
            self.delete(f"/{{id}}", summary=f"Delete {resource_class.name}")(
                self._delete
            )

    def _link_actions(self):
        resource_class = self.resource_class

        for action in inspect.getmembers(object=self):
            name, func = action

            if not isinstance(func, Action):
                continue

            for method in func.methods:
                route_method = getattr(self, method)
                response_model = (
                    self.ReadResponseModel if func.detail else self.ListResponseModel
                )
                url = f"/{{id}}/{name}" if func.detail else f"/{name}"

                route_method(
                    url,
                    summary=f"{resource_class.name} {name}",
                    response_model=response_model,
                )(func)

    def get_resource(self, request: Request):
        kwargs = self.get_resource_kwargs(request=request)
        resource = self.resource_class(**kwargs)

        setattr(request, "resource", resource)

        return resource

    def get_resource_context(self, request: Request):
        return {"request": request}

    def get_resource_kwargs(self, request: Request):
        return {"context": self.get_resource_context(request=request)}

    def build_response(
        self,
        resource: TResource,
        rows: Union[BaseModel, List[BaseModel]],
        request: Request,
    ):
        return rows

    def parse_update(
        self,
        resource: TResource,
        update: dict,
    ):
        resource_relationships = resource.get_relationships()

        relationships = {}

        for field in list(update.keys()):
            value = update[field]
            if field in resource_relationships:
                relationships[field] = value
                del update[field]

        return update, relationships

    def perform_create(
        self,
        request: Request,
        resource: TResource,
        attributes: dict,
        relationships: dict,
    ):
        if not resource.create:
            raise NotImplementedError("Resource.create not implemented")

        return resource.create(attributes=attributes, relationships=relationships)

    def perform_update(
        self,
        request: Request,
        resource: TResource,
        id: str | int,
        attributes: dict,
        relationships: dict,
    ):
        if not resource.update:
            raise NotImplementedError("Resource.update not implemented")

        return resource.update(
            id=id, attributes=attributes, relationships=relationships
        )

    def perform_delete(self, request: Request, resource: TResource, id: str | int):
        if not resource.delete:
            raise NotImplementedError("Resource.delete not implemented")

        return resource.delete(id=id)

    def _retrieve(self, *, id: Union[int, str], request: Request):
        resource = self.get_resource(request=request)
        if not resource.retrieve:
            raise NotImplementedError("Resource.retrieve not implemented")

        row = resource.retrieve(id=id)
        return self.build_response(rows=row, resource=resource, request=request)

    def _list(self, *, request: Request):
        resource = self.get_resource(request=request)
        if not resource.list:
            raise NotImplementedError("Resource.list not implemented")

        rows = resource.list()
        return self.build_response(rows=rows, resource=resource, request=request)

    def _create(self, *, create: TCreatePayload, request: Request):
        resource = self.get_resource(request=request)

        attributes, relationships = self.parse_update(
            resource=resource, update=create.dict(exclude_unset=True)
        )

        row = self.perform_create(
            request=request,
            resource=resource,
            attributes=attributes,
            relationships=relationships,
        )

        return self.build_response(rows=row, resource=resource, request=request)

    def _update(self, *, id: Union[int, str], update: TUpdatePayload, request: Request):
        resource = self.get_resource(request=request)

        attributes, relationships = self.parse_update(
            resource=resource, update=update.dict(exclude_unset=True)
        )
        row = self.perform_update(
            request=request,
            resource=resource,
            attributes=attributes,
            relationships=relationships,
            id=id,
        )

        return self.build_response(rows=row, resource=resource, request=request)

    def _delete(self, *, id: Union[int, str], request: Request):
        resource = self.get_resource(request=request)

        self.perform_delete(request=request, resource=resource, id=id)

        return Response(status_code=204)



from typing import Generic

from . import base, mixins, types

__all__ = [
    "ListCreateResource",
    "ListCreateUpdateResource",
    "SQLModelResource",
]


class ListCreateResource(
    mixins.CreateResourceMixin, mixins.ListResourceMixin, base.BaseSQLResource
):
    pass


class ListCreateUpdateResource(
    mixins.ListResourceMixin,
    mixins.CreateResourceMixin,
    mixins.UpdateResourceMixin,
    base.BaseSQLResource,
):
    pass


class SQLModelResource(
    mixins.RetrieveResourceMixin,
    mixins.ListResourceMixin,
    mixins.CreateResourceMixin,
    mixins.UpdateResourceMixin,
    mixins.DeleteResourceMixin,
    base.BaseSQLResource[types.TDb],
    types.SQLResourceProtocol[types.TDb],
    Generic[types.TDb],
):
    pass



import copy
from typing import Any, Generic, Optional, Type

from fastapi import HTTPException
from sqlalchemy.orm import MANYTOONE
from sqlalchemy.orm import exc as sa_exceptions
from sqlalchemy.orm import joinedload
from sqlmodel import Session, SQLModel, select, update
from sqlmodel.sql.expression import SelectOfScalar

from fastapi_resources.resources import base_resource
from fastapi_resources.resources.sqlmodel.exceptions import NotFound

from . import types


def get_relationships_from_schema(
    schema: Type[SQLModel],
    schema_cache: Optional[
        dict[
            tuple[Optional[str], Type[SQLModel]],
            types.SchemaWithRelationships,
        ]
    ] = None,
    immediate_parent_backpopulated_field: Optional[str] = None,
    parent_key: Optional[str] = None,
):
    schema_cache = schema_cache or {}

    relationship_fields = schema.__sqlmodel_relationships__.keys()
    relationships = {}

    parent_schema_with_relationships = types.SchemaWithRelationships(
        schema=schema,
        relationships=relationships,
    )

    # Just a sanity check. This function should never be called for a relationship
    # that already exists.
    assert (parent_key, schema) not in schema_cache

    schema_cache[(parent_key, schema)] = parent_schema_with_relationships

    for field in relationship_fields:
        sqlalchemy_relationship = getattr(schema, field).property
        related_schema = sqlalchemy_relationship.mapper.class_
        many = sqlalchemy_relationship.uselist

        # Used to uniquely identify the related schema relative to a parent
        new_parent_key = f"{schema}.{field}"

        # TODO: Handle MANYTOMANY
        direction = sqlalchemy_relationship.direction
        update_field = list(
            sqlalchemy_relationship.local_columns
            if direction == MANYTOONE
            else sqlalchemy_relationship.remote_side
        )[0].name

        # If this branch already contains the schema with the same parent,
        # we can reuse the relationship to avoid a cycle.
        if relationship_info := schema_cache.get((new_parent_key, related_schema)):
            relationships[field] = types.SQLModelRelationshipInfo(
                schema_with_relationships=relationship_info,
                many=many,
                field=field,
                direction=direction,
                update_field=update_field,
            )
            continue

        # For any relationship, the child will have a backpopulated reference
        # to the parent; we want to exclude that from available relationships
        if field == immediate_parent_backpopulated_field:
            continue

        backpopulated_field = schema.__sqlmodel_relationships__[field].back_populates

        get_relationships_from_schema(
            schema=related_schema,
            schema_cache=schema_cache,
            immediate_parent_backpopulated_field=backpopulated_field,
            parent_key=new_parent_key,
        )

        # After the above function has been called, the relationship
        # will have been created. But we want to use it without having
        # to return the object, so we just grab it from cache. A cleaner
        # solution is to overload the signature based on whether
        # `parent_field` is passed in,
        schema_with_relationship = schema_cache.get((new_parent_key, related_schema))
        assert schema_with_relationship

        relationships[field] = types.SQLModelRelationshipInfo(
            schema_with_relationships=schema_with_relationship,
            many=many,
            field=field,
            direction=direction,
            update_field=update_field,
        )

    return relationships


class BaseSQLResource(
    base_resource.Resource[types.TDb],
    types.SQLResourceProtocol[types.TDb],
    Generic[types.TDb],
):
    registry: dict[Type[SQLModel], type["BaseSQLResource"]] = {}

    def __init_subclass__(cls) -> None:
        if Db := getattr(cls, "Db", None):
            BaseSQLResource.registry[Db] = cls

        return super().__init_subclass__()

    def __init__(
        self,
        session: Session = None,
        inclusions: Optional[types.Inclusions] = None,
        *args,
        **kwargs,
    ):
        if session:
            self.session = session
        else:
            assert self.engine, "A session or an engine must be given."
            self.session = Session(self.engine)

        # TODO: Save the relationships on the instance at instantiation for caching

        super().__init__(inclusions=inclusions, *args, **kwargs)

    def close(self):
        self.session.close()

    @classmethod
    def get_relationships(cls) -> types.Relationships:
        """Builds the relationship graph for the current resource.

        TODO: Have the relationship point to the resource, rather than the schema. Use
            the registry to lookup the related resource.

        Used to:
          - Validate a given inclusion resolves to a relationship (done in the base class)
          - To retrieve all the objects along an inclusion with their schemas
        """
        read_fields = set(cls.Read.__fields__.keys())
        read_fields.update(cls.Read.__sqlmodel_relationships__.keys())
        relationships = get_relationships_from_schema(schema=cls.Db)

        return {
            name: info for name, info in relationships.items() if name in read_fields
        }

    @classmethod
    def get_attributes(cls) -> set[str]:
        attributes = set()

        # These are the fields according to the pydantic model
        fields = cls.Read.__fields__
        for field in fields:
            # If the field refers to a foreign key field we skip it as it'll be
            # included in get_relationships.
            if getattr(cls.Db, field).expressions[0].foreign_keys:
                continue

            attributes.add(field)

        return attributes

    # TODO: Update to a type from sqlalchemy when we require 2.0
    def get_where(self) -> list[Any]:
        return []

    def get_joins(self) -> list[Any]:
        return []

    def get_options(self):
        options = []
        inclusions = self.inclusions or []

        # Build the query options based on the include
        for inclusion in inclusions:
            zipped_inclusion = self.zipped_inclusions_with_resource(
                inclusion=inclusion,
            )
            option = None

            for index, zipped_field in enumerate(zipped_inclusion):
                parent = zipped_inclusion[index - 1].resource.Db if index else self.Db
                attr = getattr(parent, zipped_field.field)

                if option:
                    option = option.joinedload(attr)
                else:
                    option = joinedload(attr)

            options.append(option)

        return options

    def get_select(self):
        options = self.get_options()
        select_stmt = select(self.Db)

        for join in self.get_joins():
            select_stmt = select_stmt.join(join)

        if options := self.get_options():
            select_stmt = select_stmt.options(*options)

        if where := self.get_where():
            select_stmt = select_stmt.where(*where)

        return select_stmt

    def get_object(
        self,
        id: int | str,
    ) -> types.TDb:
        select = self.get_select()

        try:
            return self.session.exec(select.where(self.Db.id == id)).unique().one()
        except sa_exceptions.NoResultFound:
            raise NotFound(f"{self.name} not found")

    def get_related(
        self,
        obj: types.TDb,
        inclusion: list[str],
    ) -> list[types.SelectedObj]:
        """Gets related objects based on an Inclusions path."""

        def select_objs(
            _obj, _inclusion: list[str], _relationships: types.Relationships
        ) -> list[types.SelectedObj]:
            next_inclusion = copy.copy(_inclusion)
            field = next_inclusion.pop(0)
            relationship_info = _relationships[field]
            schema = relationship_info.schema_with_relationships.schema

            selected_objs = getattr(_obj, field)
            if not selected_objs:
                return []

            selected_objs = (
                selected_objs if isinstance(selected_objs, list) else [selected_objs]
            )

            selected_objs = [
                types.SelectedObj(
                    obj=selected_obj, resource=BaseSQLResource.registry[schema]
                )
                for selected_obj in selected_objs
            ]

            if next_inclusion:
                selected_objs = [
                    *selected_objs,
                    *[
                        nested_obj
                        for selected_obj in selected_objs
                        for nested_obj in select_objs(
                            _obj=selected_obj.obj,
                            _inclusion=next_inclusion,
                            _relationships=relationship_info.schema_with_relationships.relationships,
                        )
                    ],
                ]

            return selected_objs

        return select_objs(
            _obj=obj, _inclusion=inclusion, _relationships=self.get_relationships()
        )



from typing import Optional

from sqlalchemy.orm import MANYTOONE, ONETOMANY
from sqlmodel import select, update

from fastapi_resources.resources.sqlmodel import types
from fastapi_resources.resources.sqlmodel.exceptions import NotFound


class CreateResourceMixin:
    def create(
        self: types.SQLResourceProtocol[types.TDb],
        attributes: dict,
        relationships: Optional[dict[str, str | int | list[str | int]]] = None,
        **kwargs,
    ):
        row = self.Db(**attributes)

        for key, value in kwargs.items():
            setattr(row, key, value)

        relationships = relationships or {}
        model_relationships = self.get_relationships()

        for field, related_ids in relationships.items():
            relationship = model_relationships[field]
            direction = relationship.direction

            RelatedResource = self.registry[
                relationship.schema_with_relationships.schema
            ]
            related_db_model = RelatedResource.Db
            new_related_ids = (
                related_ids if isinstance(related_ids, list) else [related_ids]
            )

            # Do a select to check we have permission
            related_resource = RelatedResource(context=self.context)

            if related_where := related_resource.get_where():
                results = self.session.exec(
                    select(related_db_model).where(
                        related_db_model.id.in_(new_related_ids), *related_where
                    )
                ).all()

                if len(results) != len(new_related_ids):
                    raise NotFound()

            if direction == MANYTOONE:
                # Can update locally via a setattr
                setattr(row, relationship.update_field, new_related_ids[0])

        self.session.add(row)
        self.session.commit()

        # Update many relationships that require a separate update on the relationship table
        for field, related_ids in relationships.items():
            relationship = model_relationships[field]
            direction = relationship.direction

            if direction != ONETOMANY:
                continue

            assert isinstance(
                related_ids, list
            ), "A list of IDs must be provided for {field}"

            RelatedResource = self.registry[
                relationship.schema_with_relationships.schema
            ]
            related_db_model = RelatedResource.Db
            related_resource = RelatedResource(context=self.context)
            related_where = related_resource.get_where()

            # Update the related objects
            self.session.execute(
                update(related_db_model)
                .where(related_db_model.id.in_(related_ids), *related_where)
                .values({relationship.update_field: row.id})
            )

        return row


class UpdateResourceMixin:
    def update(
        self: types.SQLResourceProtocol[types.TDb],
        *,
        id: int | str,
        attributes: dict,
        relationships: Optional[dict[str, str | int | list[str | int]]] = None,
        **kwargs,
    ):
        row = self.get_object(id=id)

        for key, value in list(attributes.items()) + list(kwargs.items()):
            setattr(row, key, value)

        model_relationships = self.get_relationships()

        if relationships:
            for field, related_ids in relationships.items():
                relationship = model_relationships[field]
                direction = relationship.direction

                RelatedResource = self.registry[
                    relationship.schema_with_relationships.schema
                ]
                related_db_model = RelatedResource.Db
                new_related_ids = (
                    related_ids if isinstance(related_ids, list) else [related_ids]
                )

                # Do a select to check we have permission
                related_resource = RelatedResource(context=self.context)

                if related_where := related_resource.get_where():
                    results = self.session.exec(
                        select(related_db_model).where(
                            related_db_model.id.in_(new_related_ids), *related_where
                        )
                    ).all()

                    if len(results) != len(new_related_ids):
                        raise NotFound(f"{related_resource.name} not found")

                if direction == ONETOMANY:
                    assert isinstance(
                        related_ids, list
                    ), "A list of IDs must be provided for {field}"

                    # Update the related objects
                    self.session.execute(
                        update(related_db_model)
                        .where(related_db_model.id.in_(new_related_ids))
                        .values({relationship.update_field: id})
                    )

                    # Detach the old related objects
                    # NOTE: This will raise if the foreign key is required. Is this OK?
                    self.session.execute(
                        update(related_db_model)
                        .where(
                            getattr(related_db_model, relationship.update_field) == id,
                            related_db_model.id.not_in(new_related_ids),
                        )
                        .values({relationship.update_field: None})
                    )

                elif direction == MANYTOONE:
                    # Can update locally via a setattr
                    setattr(row, relationship.update_field, related_ids)

        self.session.add(row)
        self.session.commit()
        self.session.refresh(row)

        return row


class ListResourceMixin:
    def list(self: types.SQLResourceProtocol[types.TDb]):
        select = self.get_select()

        rows = self.session.exec(select).unique().all()

        return rows


class RetrieveResourceMixin:
    def retrieve(self: types.SQLResourceProtocol[types.TDb], *, id: int | str):
        row = self.get_object(id=id)

        return row


class DeleteResourceMixin:
    def delete(self: types.SQLResourceProtocol[types.TDb], *, id: int | str):
        row = self.get_object(id=id)

        self.session.delete(row)
        self.session.commit()

        return {"ok": True}



from dataclasses import dataclass
from typing import (
    Callable,
    ClassVar,
    Generic,
    List,
    Optional,
    Protocol,
    Set,
    Type,
    TypeVar,
)

from pydantic import BaseModel

Inclusions = list[list[str]]


TDb = TypeVar("TDb", bound=BaseModel)
TBaseModel = TypeVar("TBaseModel", bound=BaseModel)


@dataclass
class SchemaWithRelationships:
    schema: Type[BaseModel]
    relationships: "Relationships"


@dataclass
class RelationshipInfo:
    schema_with_relationships: SchemaWithRelationships
    many: bool
    field: str


Relationships = dict[str, RelationshipInfo]


@dataclass
class SelectedObj:
    obj: BaseModel
    resource: "ResourceProtocol"


class ResourceProtocol(Protocol, Generic[TDb, TBaseModel]):
    name: ClassVar[str]
    plural_name: ClassVar[str]

    Db: ClassVar[Type[TDb]]
    Read: ClassVar[Type[TBaseModel]]

    Create: ClassVar[Optional[Type[TBaseModel]]]
    Update: ClassVar[Optional[Type[TBaseModel]]]

    create: ClassVar[Optional[Callable]] = None
    list: ClassVar[Optional[Callable]] = None
    update: ClassVar[Optional[Callable]] = None
    delete: ClassVar[Optional[Callable]] = None
    retrieve: ClassVar[Optional[Callable]] = None

    inclusions: Inclusions
    context: dict = {}

    def __init__(*args, **kwargs):
        pass

    @classmethod
    def get_relationships(
        cls,
    ) -> dict[str, RelationshipInfo]:
        ...

    @classmethod
    def get_attributes(cls) -> Set[str]:
        """Get the non-relationships attributes for the resource.

        The attributes that end up in the response depend on those specified in
        the Read schema; it'll be a subset of those returned here.
        """
        ...

    def get_related(self, obj: BaseModel, inclusion: List[str]) -> List[SelectedObj]:
        ...



from copy import copy
from dataclasses import dataclass
from typing import ClassVar, Generic, Optional, Type

from pydantic.main import BaseModel

from .types import Inclusions, Relationships, SelectedObj, TDb


@dataclass
class InclusionWithResource:
    field: str
    resource: type["Resource"]


class Resource(Generic[TDb]):
    # name: ClassVar[str]
    plural_name: ClassVar[Optional[str]]

    Db: ClassVar[Type[TDb]]
    registry: dict[Type[BaseModel], type["Resource"]] = {}

    def __init_subclass__(cls) -> None:
        if Db := getattr(cls, "Db", None):
            Resource.registry[Db] = cls

        # Pluralize name
        if name := getattr(cls, "name", None):
            cls.plural_name = getattr(cls, "plural_name", None) or f"{name}s"

        return super().__init_subclass__()

    def __init__(
        self,
        inclusions: Optional[Inclusions] = None,
        context: Optional[dict] = None,
        *args,
        **kwargs,
    ):
        if inclusions:
            self.validate_inclusions(inclusions=inclusions)

        self.inclusions = inclusions or []
        self.context = context or {}

    def close(self):
        pass

    def _zipped_inclusions_with_resource(
        self, _relationships: Relationships, _inclusion: list[str]
    ) -> list[InclusionWithResource]:
        if not _inclusion:
            return []

        next_inclusions = copy(_inclusion)
        current_inclusion = next_inclusions.pop(0)

        relationship_info = _relationships.get(current_inclusion)
        assert relationship_info, f"Invalid inclusion {current_inclusion}"

        if relationship_info.schema_with_relationships.schema not in self.registry:
            raise Exception(
                f"Resource not found for relationship {relationship_info.field}"
            )

        resource = self.registry[relationship_info.schema_with_relationships.schema]

        zipped_inclusions = [
            InclusionWithResource(field=current_inclusion, resource=resource)
        ]

        if (
            next_relationships := relationship_info.schema_with_relationships.relationships
        ):
            zipped_inclusions = [
                *zipped_inclusions,
                *self._zipped_inclusions_with_resource(
                    _relationships=next_relationships,
                    _inclusion=next_inclusions,
                ),
            ]

        return zipped_inclusions

    def validate_inclusions(self, inclusions: Inclusions):
        """Validate the inclusions by walking the relationships."""
        relationships = self.get_relationships()

        for inclusion in inclusions:
            self._zipped_inclusions_with_resource(
                _relationships=relationships, _inclusion=inclusion
            )

    def zipped_inclusions_with_resource(
        self, inclusion: list[str]
    ) -> list[InclusionWithResource]:
        return self._zipped_inclusions_with_resource(
            _relationships=self.get_relationships(),
            _inclusion=inclusion,
        )

    @classmethod
    def get_relationships(cls) -> Relationships:
        return {}

    @classmethod
    def get_attributes(cls) -> set[str]:
        return set(cls.Db.__fields__.keys())

    def get_related(self, obj: BaseModel, inclusion: list[str]) -> list[SelectedObj]:
        raise NotImplementedError()



import functools
from typing import Callable, Literal

methods = Literal["post", "get", "patch", "delete"]


def action(detail: bool, methods: list[methods] = ["get"], **kwargs):
    def action_decorator(func: Callable):
        func.detail = detail
        func.methods = methods
        func.kwargs = kwargs

        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            self = args[0]
            request = kwargs["request"]
            resource = self.get_resource(request=request)

            rows = func(*args, **kwargs)

            return self.build_response(resource=resource, rows=rows, request=request)

        return wrapped

    return action_decorator



from dataclasses import dataclass
from typing import ClassVar, Generic, Optional, Protocol, Type, TypeVar

from sqlalchemy.engine import Engine
from sqlalchemy.orm import MANYTOMANY, ONETOMANY
from sqlmodel import Session, SQLModel
from sqlmodel.sql.expression import SelectOfScalar

from fastapi_resources.resources import types

Inclusions = types.Inclusions


@dataclass()
class SchemaWithRelationships:
    schema: Type[SQLModel]
    relationships: "Relationships"


@dataclass
class SQLModelRelationshipInfo:
    schema_with_relationships: SchemaWithRelationships
    many: bool
    field: str
    direction: ONETOMANY | MANYTOMANY
    update_field: str


Relationships = dict[str, SQLModelRelationshipInfo]


TDb = TypeVar("TDb", bound=SQLModel)


class SQLResourceProtocol(types.ResourceProtocol, Protocol, Generic[TDb]):
    Db: ClassVar[Type[TDb]]
    Read: ClassVar[Type[SQLModel]]
    Create: ClassVar[Optional[Type[SQLModel]]] = None
    Update: ClassVar[Optional[Type[SQLModel]]] = None

    engine: ClassVar[Optional[Engine]] = None

    session: Session

    registry: dict[Type[SQLModel], type["SQLResourceProtocol"]] = {}

    @classmethod
    def get_relationships(
        cls,
    ) -> dict[str, SQLModelRelationshipInfo]:
        """Get the relationships for the resource."""
        ...

    @classmethod
    def get_attributes(cls) -> set[str]:
        """Get the non-relationships attributes for the resource.

        The attributes that end up in the response depend on those specified in
        the Read schema; it'll be a subset of those returned here.
        """
        ...

    def get_related(self, obj: SQLModel, inclusion: list[str]) -> list[TDb]:
        ...

    def get_object(self, id: int | str) -> TDb:
        ...

    def get_select(self) -> SelectOfScalar[TDb]:
        ...

    def get_where(self) -> list[str]:
        ...


@dataclass
class SelectedObj:
    obj: SQLModel
    resource: "SQLResourceProtocol"



from typing import List, Literal, Optional, Type, TypeVar, Union

from fastapi import Query, Request
from pydantic import create_model
from pydantic.main import BaseModel, ModelMetaclass

from fastapi_resources.resources.types import (
    Inclusions,
    RelationshipInfo,
    Relationships,
    ResourceProtocol,
)
from fastapi_resources.routers import base_router

from . import types

include_query = Query(None, regex=r"^([\w\.]+)(,[\w\.]+)*$")


TResource = TypeVar("TResource", bound=ResourceProtocol)


def get_schemas_from_relationships(
    relationships: Relationships, visited: Optional[set[type[types.Object]]] = None
) -> list[tuple[str, ModelMetaclass]]:
    schemas: list[tuple[str, ModelMetaclass]] = []

    visited = visited or set()
    for relationship_info in relationships.values():
        schema = relationship_info.schema_with_relationships.schema
        if schema in visited:
            continue

        visited.add(schema)
        schemas.append((relationship_info.field, schema))
        schemas += get_schemas_from_relationships(
            relationships=relationship_info.schema_with_relationships.relationships,
            visited=visited,
        )

    return schemas


def get_relationships_model_for_model(
    method: str, resource_class: type[TResource], model: type[BaseModel]
):
    allowed_fields = set(model.__fields__.keys())
    allowed_fields.update(model.__sqlmodel_relationships__.keys())

    RelationshipLinkages = {
        relationship_name: (
            Optional[
                create_model(
                    f"{model.__name__}__{method}__Relationships{relationship_name}",
                    __base__=(
                        (
                            types.JARelationshipsObjectMany
                            if relationship_info.many
                            else types.JARelationshipsObjectSingle
                        )
                    )[
                        Literal[
                            resource_class.registry[
                                relationship_info.schema_with_relationships.schema
                            ].name  # type: ignore
                        ]
                    ],
                )
            ],
            None,
        )
        for relationship_name, relationship_info in resource_class.get_relationships().items()
        if relationship_info.schema_with_relationships.schema in resource_class.registry
        and relationship_info.field in allowed_fields
    }

    Relationships = create_model(
        f"{model.__name__}__{method}__Relationships",
        **RelationshipLinkages,
        __base__=BaseModel,
    )

    return Relationships


def get_attributes_model_for_model(
    method: str, model: type[BaseModel], resource_class: type[TResource]
):
    Attributes = create_model(f"{model.__name__}__{method}__Attributes", __base__=model)

    # Remove the ID
    if "id" in Attributes.__fields__:
        del Attributes.__fields__["id"]

    return Attributes


def get_model_for_resource_class(
    method: str, model: type[BaseModel], resource_class: type[TResource]
):
    Attributes = get_attributes_model_for_model(
        method=method, model=model, resource_class=resource_class
    )
    Relationships = get_relationships_model_for_model(
        method=method, model=model, resource_class=resource_class
    )

    return types.JAResourceObject[
        Attributes,
        Relationships,
        Literal[(resource_class.name,)],  # type: ignore
    ]


class JSONAPIResourceRouter(base_router.ResourceRouter[TResource]):
    def __init__(
        self,
        *,
        resource_class: type[TResource],
        **kwargs,
    ) -> None:
        self.resource_class = resource_class

        super().__init__(
            resource_class=resource_class,
            prefix=f"/{resource_class.plural_name}",
            **kwargs,
        )

    def get_included_schema(
        self, method: str
    ) -> tuple[type[types.JAResourceObject], ...]:
        relationships = self.resource_class.get_relationships()
        schemas = get_schemas_from_relationships(relationships=relationships)

        return tuple(
            get_model_for_resource_class(
                method=f"{self.prefix}__{method}__included__{field}__{schema.__name__}",
                resource_class=self.resource_class.registry[schema],
                model=self.resource_class.registry[schema].Read,
            )
            for field, schema in schemas
            if schema in self.resource_class.registry
        )

    def get_read_response_model(self):
        included_schemas = self.get_included_schema(method="retrieve")
        Included = List[Union[included_schemas]] if included_schemas else list  # type: ignore
        Name = Literal[(self.resource_class.name,)]  # type: ignore

        Attributes = get_attributes_model_for_model(
            method="retrieve",
            model=self.resource_class.Read,
            resource_class=self.resource_class,
        )
        Relationships = get_relationships_model_for_model(
            method="retrieve",
            model=self.resource_class.Read,
            resource_class=self.resource_class,
        )

        return types.JAResponseSingle[Attributes, Relationships, Name, Included]

    def get_list_response_model(self):
        included_schemas = self.get_included_schema(method="list")
        Included = List[Union[included_schemas]] if included_schemas else list  # type: ignore
        Name = Literal[(self.resource_class.name,)]  # type: ignore

        Attributes = get_attributes_model_for_model(
            method="list",
            model=self.resource_class.Read,
            resource_class=self.resource_class,
        )
        Relationships = get_relationships_model_for_model(
            method="list",
            model=self.resource_class.Read,
            resource_class=self.resource_class,
        )

        return types.JAResponseList[Attributes, Relationships, Name, Included]

    def get_update_model(self):
        if not self.resource_class.Update:
            return None

        Name = Literal[(self.resource_class.name,)]  # type: ignore

        Attributes = get_attributes_model_for_model(
            method="update",
            model=self.resource_class.Update,
            resource_class=self.resource_class,
        )
        Relationships = get_relationships_model_for_model(
            method="update",
            model=self.resource_class.Update,
            resource_class=self.resource_class,
        )

        return types.JAUpdateRequest[Attributes, Relationships, Name]

    def get_create_model(self):
        if not self.resource_class.Create:
            return None

        Name = Literal[(self.resource_class.name,)]  # type: ignore

        Attributes = get_attributes_model_for_model(
            method="create",
            model=self.resource_class.Create,
            resource_class=self.resource_class,
        )
        Relationships = get_relationships_model_for_model(
            method="create",
            model=self.resource_class.Create,
            resource_class=self.resource_class,
        )

        return types.JACreateRequest[Attributes, Relationships, Name]

    def get_resource_kwargs(self, request: Request):
        inclusions: Inclusions = []
        include = request.query_params.get("include")

        if include:
            inclusions = [inclusion.split(".") for inclusion in include.split(",")]

        for relationship in self.resource_class.get_relationships().values():
            inclusions.append([relationship.field])

        return {
            **super().get_resource_kwargs(request=request),
            "inclusions": inclusions,
        }

    def build_document_links(self, request: Request):
        path = request.url.path

        if query := request.url.query:
            path = f"{path}?{query}"

        return types.JALinks(self=path)

    def build_resource_object_links(
        self, id: str, resource: Union[Type[TResource], TResource]
    ):
        return types.JALinks(self=f"/{resource.plural_name}/{id}")

    def build_resource_identifier_object(
        self,
        related_obj: types.Object,
        resource: Union[Type[TResource], TResource],
        relationship_info: RelationshipInfo,
    ) -> types.JAResourceIdentifierObject:
        return types.JAResourceIdentifierObject(
            type=resource.registry[
                relationship_info.schema_with_relationships.schema
            ].name,
            id=str(related_obj.id),
        )

    def build_resource_object_relationships(
        self, obj: types.Object, resource: TResource
    ):
        relationships = {}

        for (
            relationship_name,
            relationship_info,
        ) in resource.get_relationships().items():
            links = types.JALinks(
                self=f"/{resource.plural_name}/{obj.id}/relationships/{relationship_name}",
                related=f"/{resource.plural_name}/{obj.id}/{relationship_name}",
            )

            # The relationships will have been properly selected, so this should not send
            # another query.
            data = [
                self.build_resource_identifier_object(
                    related_obj=related_obj.obj,
                    resource=resource,
                    relationship_info=relationship_info,
                )
                for related_obj in resource.get_related(obj, [relationship_name])
                if related_obj.obj
            ]

            if relationship_info.many:
                relationship_object = types.JARelationshipsObjectMany(
                    links=links,
                    data=data,
                )
            else:
                data = data[0] if data else None
                relationship_object = types.JARelationshipsObjectSingle(
                    links=links,
                    data=data,
                )

            relationships[relationship_name] = relationship_object

        # If the relationship is to-one, then we can include `data`
        # But for to-many, we only include it if it's in inclusions.
        return relationships

    def build_resource_object(self, obj: types.Object, resource: TResource):
        valid_attributes = resource.get_attributes()

        # ID is a special case, so can ignored
        valid_attributes.remove("id")

        # Filter out relationships attributes
        attributes = {
            key: value
            for key, value in resource.Read.from_orm(obj).dict().items()
            if key in valid_attributes
        }

        resource_object = types.JAResourceObject(
            id=str(obj.id),
            type=resource.name,
            attributes=attributes,
            links=self.build_resource_object_links(id=obj.id, resource=resource),
            relationships=self.build_resource_object_relationships(
                obj=obj, resource=resource
            ),
        )

        return resource_object

    def build_response(
        self,
        rows: Union[types.Object, list[types.Object]],
        resource: TResource,
        request: Request,
    ):
        included_resources = {}

        many = isinstance(rows, list)
        rows = rows if isinstance(rows, list) else [rows]

        include = request.query_params.get("include")
        inclusions = []

        if include:
            inclusions = [inclusion.split(".") for inclusion in include.split(",")]

        for row in rows:
            for inclusion in inclusions:
                selected_objs = resource.get_related(obj=row, inclusion=inclusion)

                for selected_obj in selected_objs:
                    obj = selected_obj.obj
                    related_resource = selected_obj.resource

                    included_resources[
                        (related_resource.name, obj.id)
                    ] = self.build_resource_object(obj=obj, resource=related_resource())

        data = [self.build_resource_object(obj=row, resource=resource) for row in rows]
        included = list(included_resources.values())

        # Get top-level resource links
        links = self.build_document_links(request=request)

        if many:
            return types.JAResponseList(data=data, included=included, links=links)

        return types.JAResponseSingle(data=data[0], included=included, links=links)

    def _parse_request_payload(self, payload: dict):
        # Merge the attributes and relationships into a single update
        parsed_relationships = {}

        for key, linkage in payload["data"].get("relationships", {}).items():
            identifiers = linkage["data"]
            if isinstance(identifiers, list):
                parsed_relationships[key] = [ro["id"] for ro in identifiers]
            else:
                parsed_relationships[key] = identifiers["id"]

        merged_payload = {
            **payload["data"].get("attributes", {}),
            **parsed_relationships,
        }

        return merged_payload

    def parse_update(self, resource: TResource, update: dict):
        return super().parse_update(
            resource, self._parse_request_payload(payload=update)
        )

    def parse_create(self, resource: TResource, create: dict):
        return super().parse_update(
            resource, self._parse_request_payload(payload=create)
        )

    def _retrieve(
        self,
        *,
        id: Union[int, str],
        request: Request,
        include: Optional[str] = include_query,
    ):
        return super()._retrieve(id=id, request=request)

    def _list(self, *, request: Request, include: Optional[str] = include_query):
        return super()._list(request=request)

    def _create(
        self,
        *,
        create: base_router.TCreatePayload,
        request: Request,
        include: Optional[str] = include_query,
    ):
        return super()._create(create=create, request=request)

    def _update(
        self,
        *,
        id: Union[int, str],
        update: base_router.TUpdatePayload,
        request: Request,
        include: Optional[str] = include_query,
    ):
        return super()._update(id=id, update=update, request=request)

    def _delete(self, *, id: Union[int, str], request: Request):
        return super()._delete(id=id, request=request)



from typing import Generic, List, Optional, TypeVar

from pydantic import BaseModel
from pydantic.generics import GenericModel


class Object(BaseModel):
    id: str


TRead = TypeVar("TRead", bound=Object)
TName = TypeVar("TName", bound=str)
TUpdate = TypeVar("TUpdate", bound=Object)
TCreate = TypeVar("TCreate", bound=Object)
TAttributes = TypeVar("TAttributes")
TRelationships = TypeVar("TRelationships")
TIncluded = TypeVar("TIncluded")
TType = TypeVar("TType", bound=str)


class TIncludeParam(str):
    pass


class JALinks(BaseModel):
    """A links-object"""

    self: Optional[str] = ""
    # Will be used when relationship endpoints are implemented
    related: Optional[str] = ""


class JAResourceIdentifierObject(GenericModel, Generic[TType]):
    type: TType
    id: str


class JARelationshipsObjectSingle(GenericModel, Generic[TType]):
    links: Optional[JALinks]
    data: Optional[JAResourceIdentifierObject[TType]] = None


class JARelationshipsObjectMany(GenericModel, Generic[TType]):
    links: Optional[JALinks]
    data: list[JAResourceIdentifierObject[TType]]


class JAResourceObject(GenericModel, Generic[TAttributes, TRelationships, TName]):
    id: str
    type: TName
    attributes: TAttributes
    links: JALinks
    relationships: TRelationships


class JAUpdateObject(GenericModel, Generic[TAttributes, TRelationships, TName]):
    id: str
    type: TName
    attributes: Optional[TAttributes]
    relationships: Optional[TRelationships]


class JAUpdateRequest(GenericModel, Generic[TAttributes, TRelationships, TName]):
    data: JAUpdateObject[TAttributes, TRelationships, TName]


class JACreateObject(GenericModel, Generic[TAttributes, TRelationships, TName]):
    type: TName
    attributes: Optional[TAttributes]
    relationships: Optional[TRelationships]


class JACreateRequest(GenericModel, Generic[TAttributes, TRelationships, TName]):
    data: JACreateObject[TAttributes, TRelationships, TName]


class JAResponseSingle(
    GenericModel, Generic[TAttributes, TRelationships, TName, TIncluded]
):
    data: JAResourceObject[TAttributes, TRelationships, TName]
    included: TIncluded
    links: JALinks


class JAResponseList(
    GenericModel, Generic[TAttributes, TRelationships, TName, TIncluded]
):
    data: List[JAResourceObject[TAttributes, TRelationships, TName]]
    included: TIncluded
    links: JALinks



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fastapi_resources',
 'fastapi_resources.resources',
 'fastapi_resources.resources.sqlmodel',
 'fastapi_resources.routers',
 'fastapi_resources.routers.json_api_router']

package_data = \
{'': ['*']}

install_requires = \
['fastapi>=0.85.0,<0.86.0', 'sqlmodel>=0.0.8,<0.0.9']

setup_kwargs = {
    'name': 'fastapi-resources',
    'version': '0.2.15',
    'description': '',
    'long_description': 'None',
    'author': 'Ben Davis',
    'author_email': 'ben@bencdavis.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
